@@ -4,11 +4,13 @@ import bisect
 import collections
 import copy
 import datetime
-import functools
 import itertools
 import math
 import numbers
 import random
+import warnings
+
+from sentinels import NOTHING
 import six
 from six import moves
 
@@ -86,6 +88,7 @@ string_operators = [
     '$substr',
     '$toLower',
     '$toUpper',
+    '$toString',
 ]
 comparison_operators = [
     '$cmp',
@@ -130,8 +133,9 @@ _GROUPING_OPERATOR_MAP = {
 class _Parser(object):
     """Helper to parse expressions within the aggregate pipeline."""
 
-    def __init__(self, doc_dict, user_vars=None):
+    def __init__(self, doc_dict, user_vars=None, ignore_missing_keys=False):
         self._doc_dict = doc_dict
+        self._ignore_missing_keys = ignore_missing_keys
         self._user_vars = user_vars or {}
 
     def parse(self, expression):
@@ -157,13 +161,21 @@ class _Parser(object):
                 return self._handle_conditional_operator(k, v)
             if k in set_operators:
                 return self._handle_set_operator(k, v)
-            if k in boolean_operators + string_operators + \
+            if k in string_operators:
+                return self._handle_string_operator(k, v)
+            if k in boolean_operators + \
                     text_search_operators + projection_operators:
                 raise NotImplementedError(
                     "'%s' is a valid operation but it is not supported by Mongomock yet." % k)
             if k.startswith('$'):
                 raise OperationFailure("Unrecognized expression '%s'" % k)
-            value_dict[k] = self.parse(v)
+            try:
+                value = self.parse(v)
+            except KeyError:
+                if self._ignore_missing_keys:
+                    continue
+                raise
+            value_dict[k] = value
 
         return value_dict
 
@@ -256,6 +268,46 @@ class _Parser(object):
             'aggregation pipeline, it is currently not implemented '
             ' in Mongomock.' % operator)
 
+    def _handle_string_operator(self, operator, values):
+        if operator == '$toLower':
+            parsed = self.parse(values)
+            return str(parsed).lower() if parsed is not None else ''
+        if operator == '$toUpper':
+            parsed = self.parse(values)
+            return str(parsed).upper() if parsed is not None else ''
+        if operator == '$concat':
+            parsed_list = [self.parse(value) for value in values]
+            return None if None in parsed_list else ''.join([str(x) for x in parsed_list])
+        if operator == '$substr':
+            if len(values) != 3:
+                raise OperationFailure('substr must have 3 items')
+            string = str(self.parse(values[0]))
+            first = self.parse(values[1])
+            length = self.parse(values[2])
+            if string is None:
+                return ''
+            if first < 0:
+                warnings.warn('Negative starting point given to $substr is accepted only until '
+                              'MongoDB 3.7. This behavior will change in the future.')
+                return ''
+            if length < 0:
+                warnings.warn('Negative length given to $substr is accepted only until '
+                              'MongoDB 3.7. This behavior will change in the future.')
+            second = len(string) if length < 0 else first + length
+            return string[first:second]
+        if operator == '$strcasecmp':
+            if len(values) != 2:
+                raise OperationFailure('strcasecmp must have 2 items')
+            a, b = str(self.parse(values[0])), str(self.parse(values[1]))
+            return 0 if a == b else -1 if a < b else 1
+        if operator == '$toString':
+            parsed = self.parse(values)
+            return str(parsed) if parsed is not None else None
+        # This should never happen: it is only a safe fallback if something went wrong.
+        raise NotImplementedError(  # pragma: no cover
+            "Although '%s' is a valid string operator for the aggregation "
+            'pipeline, it is currently not implemented  in Mongomock.' % operator)
+
     def _handle_date_operator(self, operator, values):
         out_value = self.parse(values)
         if operator == '$dayOfYear':
@@ -311,7 +363,11 @@ class _Parser(object):
             cond = value['cond']
             return [
                 item for item in input_array
-                if _Parser(self._doc_dict, dict(self._user_vars, **{fieldname: item})).parse(cond)
+                if _Parser(
+                    self._doc_dict,
+                    dict(self._user_vars, **{fieldname: item}),
+                    ignore_missing_keys=self._ignore_missing_keys,
+                ).parse(cond)
             ]
 
         raise NotImplementedError(
@@ -364,9 +420,17 @@ class _Parser(object):
             'pipeline, it is currently not implemented in Mongomock.' % operator)
 
 
-def _parse_expression(expression, doc_dict):
-    """Parse an expression."""
-    return _Parser(doc_dict).parse(expression)
+def _parse_expression(expression, doc_dict, ignore_missing_keys=False):
+    """Parse an expression.
+
+    Args:
+        expression: an Aggregate Expression, see
+            https://docs.mongodb.com/manual/meta/aggregation-quick-reference/#aggregation-expressions.
+        doc_dict: the document on which to evaluate the expression.
+        ignore_missing_keys: if True, missing keys evaluated by the expression are ignored silently
+            if it is possible.
+    """
+    return _Parser(doc_dict, ignore_missing_keys=ignore_missing_keys).parse(expression)
 
 
 def _accumulate_group(output_fields, group_list):
@@ -375,8 +439,12 @@ def _accumulate_group(output_fields, group_list):
         if field == '_id':
             continue
         for operator, key in six.iteritems(value):
-            key_getter = functools.partial(_parse_expression, key)
-            values = [key_getter(doc) for doc in group_list]
+            values = []
+            for doc in group_list:
+                try:
+                    values.append(_parse_expression(key, doc))
+                except KeyError:
+                    continue
             if operator in _GROUPING_OPERATOR_MAP:
                 doc_dict[field] = _GROUPING_OPERATOR_MAP[operator](values)
             elif operator == '$first':
@@ -455,12 +523,81 @@ def _handle_lookup_stage(in_collection, database, options):
         if isinstance(query, list):
             query = {'$in': query}
         matches = foreign_collection.find({foreign_field: query})
-        # TODO(pascal): Do not modify the input collection.
         doc[local_name] = [foreign_doc for foreign_doc in matches]
 
     return in_collection
 
 
+def _handle_graph_lookup_stage(in_collection, database, options):
+    if not isinstance(options.get('maxDepth', 0), six.integer_types):
+        raise OperationFailure(
+            "Argument 'maxDepth' to $graphLookup must be a number")
+    if not isinstance(options.get('restrictSearchWithMatch', {}), dict):
+        raise OperationFailure(
+            "Argument 'restrictSearchWithMatch' to $graphLookup must be a Dictionary")
+    if not isinstance(options.get('depthField', ''), six.string_types):
+        raise OperationFailure(
+            "Argument 'depthField' to $graphlookup must be a string")
+    if 'startWith' not in options:
+        raise OperationFailure(
+            "Must specify 'startWith' field for a $graphLookup")
+    for operator in ('as', 'connectFromField', 'connectToField', 'from'):
+        if operator not in options:
+            raise OperationFailure(
+                "Must specify '%s' field for a $graphLookup" % operator)
+        if not isinstance(options[operator], six.string_types):
+            raise OperationFailure(
+                "Argument '%s' to $graphLookup must be string" % operator)
+        if options[operator].startswith('$'):
+            raise OperationFailure("FieldPath field names may not start with '$'")
+        if operator in ('connectFromField', 'as') and \
+                '.' in options[operator]:
+            raise NotImplementedError(
+                "Although '.' is valid in the '%s' "
+                'parameter for the $graphLookup stage of the aggregation '
+                'pipeline, it is currently not implemented in Mongomock.' % operator)
+
+    foreign_name = options['from']
+    start_with = options['startWith']
+    connect_from_field = options['connectFromField']
+    connect_to_field = options['connectToField']
+    local_name = options['as']
+    max_depth = options.get('maxDepth', None)
+    depth_field = options.get('depthField', None)
+    restrict_search_with_match = options.get('restrictSearchWithMatch', {})
+    foreign_collection = database.get_collection(foreign_name)
+    out_doc = copy.deepcopy(in_collection)  # TODO(pascal): speed the deep copy
+
+    def _find_matches_for_depth(query):
+        if isinstance(query, list):
+            query = {'$in': query}
+        matches = foreign_collection.find({connect_to_field: query})
+        new_matches = []
+        for new_match in matches:
+            if filtering.filter_applies(restrict_search_with_match, new_match) \
+                    and new_match['_id'] not in found_items:
+                if depth_field is not None:
+                    new_match = collections.OrderedDict(new_match, **{depth_field: depth})
+                new_matches.append(new_match)
+                found_items.add(new_match['_id'])
+        return new_matches
+
+    for doc in out_doc:
+        found_items = set()
+        depth = 0
+        result = _parse_expression(start_with, doc)
+        origin_matches = doc[local_name] = _find_matches_for_depth(result)
+        while origin_matches and (max_depth is None or depth < max_depth):
+            depth += 1
+            newly_discovered_matches = []
+            for match in origin_matches:
+                match_target = match.get(connect_from_field)
+                newly_discovered_matches += _find_matches_for_depth(match_target)
+            doc[local_name] += newly_discovered_matches
+            origin_matches = newly_discovered_matches
+    return out_doc
+
+
 def _handle_group_stage(in_collection, unused_database, options):
     grouped_collection = []
     _id = options['_id']
@@ -682,6 +819,24 @@ def _project_by_spec(doc, proj_spec, is_include):
     return output
 
 
+def _handle_replace_root_stage(in_collection, unused_database, options):
+    if 'newRoot' not in options:
+        raise OperationFailure("Parameter 'newRoot' is missing for $replaceRoot operation.")
+    new_root = options['newRoot']
+    out_collection = []
+    for doc in in_collection:
+        try:
+            new_doc = _parse_expression(new_root, doc, ignore_missing_keys=True)
+        except KeyError:
+            new_doc = NOTHING
+        if not isinstance(new_doc, dict):
+            raise OperationFailure(
+                "'newRoot' expression must evaluate to an object, but resulting value was: {}"
+                .format(new_doc))
+        out_collection.append(new_doc)
+    return out_collection
+
+
 def _handle_project_stage(in_collection, unused_database, options):
     filter_list = []
     method = None
@@ -738,7 +893,8 @@ def _handle_out_stage(in_collection, database, options):
     out_collection = database.get_collection(options)
     if out_collection.count() > 0:
         out_collection.drop()
-    out_collection.insert_many(in_collection)
+    if in_collection:
+        out_collection.insert_many(in_collection)
     return in_collection
 
 
@@ -761,7 +917,7 @@ _PIPELINE_HANDLERS = {
     '$currentOp': None,
     '$facet': None,
     '$geoNear': None,
-    '$graphLookup': None,
+    '$graphLookup': _handle_graph_lookup_stage,
     '$group': _handle_group_stage,
     '$indexStats': None,
     '$limit': lambda c, d, o: c[:o],
@@ -772,7 +928,7 @@ _PIPELINE_HANDLERS = {
     '$out': _handle_out_stage,
     '$project': _handle_project_stage,
     '$redact': None,
-    '$replaceRoot': None,
+    '$replaceRoot': _handle_replace_root_stage,
     '$sample': _handle_sample_stage,
     '$skip': lambda c, d, o: c[o:],
     '$sort': _handle_sort_stage,
