@@ -16,8 +16,11 @@ import warnings
 
 try:
     from bson import json_util, SON, BSON, Timestamp
+    from bson import codec_options as bson_codec_options
+    _DEFAULT_CODEC_OPTIONS = bson_codec_options.CodecOptions()
 except ImportError:
-    json_utils = SON = BSON = Timestamp = None
+    bson_codec_options = json_utils = SON = BSON = Timestamp = None
+    _DEFAULT_CODEC_OPTIONS = None
 try:
     import execjs
 except ImportError:
@@ -67,7 +70,7 @@ _KwargOption = collections.namedtuple('KwargOption', ['typename', 'default', 'at
 
 _WITH_OPTIONS_KWARGS = {
     'codec_options': _KwargOption(
-        'bson.codec_options.CodecOptions', None,
+        'bson.codec_options.CodecOptions', _DEFAULT_CODEC_OPTIONS,
         ('document_class', 'tz_aware', 'uuid_representation')),
     'read_preference': _KwargOption(
         'pymongo.read_preference.ReadPreference', _READ_PREFERENCE_PRIMARY,
@@ -263,7 +266,7 @@ class BulkOperationBuilder(object):
 
     def insert(self, doc):
         def exec_insert():
-            self.collection.insert(doc)
+            self.collection.insert_one(doc)
             return {'nInserted': 1}
         self.executors.append(exec_insert)
 
@@ -359,9 +362,10 @@ class BulkOperationBuilder(object):
 
 class Collection(object):
 
-    def __init__(self, database, _store, write_concern=None, read_preference=None):
+    def __init__(self, database, name, _db_store, write_concern=None, read_preference=None):
         self.database = database
-        self._store = _store
+        self._name = name
+        self._db_store = _db_store
         self._write_concern = write_concern or WriteConcern()
         self._read_preference = read_preference or _READ_PREFERENCE_PRIMARY
 
@@ -380,11 +384,11 @@ class Collection(object):
 
     @property
     def full_name(self):
-        return '{0}.{1}'.format(self.database.name, self._store.name)
+        return '{0}.{1}'.format(self.database.name, self._name)
 
     @property
     def name(self):
-        return self._store.name
+        return self._name
 
     @property
     def write_concern(self):
@@ -394,6 +398,14 @@ class Collection(object):
     def read_preference(self):
         return self._read_preference
 
+    @property
+    def codec_options(self):
+        if not bson_codec_options:
+            raise NotImplementedError(
+                'The codec options are not implemented in mongomock alone, you need to import '
+                'the pymongo library as well.')
+        return bson_codec_options.CodecOptions()
+
     def initialize_unordered_bulk_op(self):
         return BulkOperationBuilder(self, ordered=False)
 
@@ -417,26 +429,41 @@ class Collection(object):
         documents = list(documents)
         for document in documents:
             validate_is_mutable_mapping('document', document)
-        return InsertManyResult(self._insert(documents, session), acknowledged=True)
+        return InsertManyResult(
+            self._insert(documents, session, ordered=ordered),
+            acknowledged=True)
 
-    def _insert(self, data, session=None):
+    @property
+    def _store(self):
+        return self._db_store[self._name]
+
+    def _insert(self, data, session=None, ordered=True):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
         if not isinstance(data, Mapping):
             results = []
+            write_errors = []
+            num_inserted = 0
             for index, item in enumerate(data):
                 try:
                     results.append(self._insert(item))
                 except WriteError as error:
-                    raise BulkWriteError({
-                        'writeErrors': [{
-                            'index': index,
-                            'code': error.code,
-                            'errmsg': str(error),
-                            'op': item,
-                        }],
-                        'nInserted': index,
+                    write_errors.append({
+                        'index': index,
+                        'code': error.code,
+                        'errmsg': str(error),
+                        'op': item,
                     })
+                    if ordered:
+                        break
+                    else:
+                        continue
+                num_inserted += 1
+            if write_errors:
+                raise BulkWriteError({
+                    'writeErrors': write_errors,
+                    'nInserted': num_inserted,
+                })
             return results
 
         # Like pymongo, we should fill the _id in the inserted dict (odd behavior,
@@ -665,7 +692,15 @@ class Collection(object):
                             arr_copy = copy.deepcopy(arr)
                             if isinstance(value, dict):
                                 for obj in arr_copy:
-                                    if filter_applies(value, obj):
+                                    try:
+                                        is_matching = filter_applies(value, obj)
+                                    except OperationFailure:
+                                        is_matching = False
+                                    if is_matching:
+                                        arr.remove(obj)
+                                        continue
+
+                                    if filter_applies({field: value}, {field: obj}):
                                         arr.remove(obj)
                             else:
                                 for obj in arr_copy:
@@ -907,12 +942,14 @@ class Collection(object):
     def find(self, filter=None, projection=None, skip=0, limit=0,
              no_cursor_timeout=False, cursor_type=None, sort=None,
              allow_partial_results=False, oplog_replay=False, modifiers=None,
-             batch_size=0, manipulate=True, collation=None, session=None):
+             batch_size=0, manipulate=True, collation=None, session=None,
+             max_time_ms=None):
         spec = filter
         if spec is None:
             spec = {}
         validate_is_mapping('filter', spec)
-        return Cursor(self, spec, sort, projection, skip, limit, collation=collation)
+        return Cursor(self, spec, sort, projection, skip, limit,
+                      collation=collation).max_time_ms(max_time_ms)
 
     def _get_dataset(self, spec, sort, fields, as_class):
         dataset = self._iter_documents(spec)
@@ -922,6 +959,9 @@ class Collection(object):
                     if sort_direction < 0:
                         dataset = iter(reversed(list(dataset)))
                     continue
+                if sort_key.startswith('$'):
+                    raise NotImplementedError(
+                        'Sorting by {} is not implemented in mongomock yet'.format(sort_key))
                 dataset = iter(sorted(
                     dataset, key=lambda x: resolve_sort_key(sort_key, x),
                     reverse=sort_direction < 0))
@@ -1587,16 +1627,12 @@ class Collection(object):
         in_collection = [doc for doc in self.find()]
         return aggregate.process_pipeline(in_collection, self.database, pipeline, session)
 
-    def with_options(self, **kwargs):
-        forbidden_kwargs = set(kwargs.keys()) - set(_WITH_OPTIONS_KWARGS)
-        if forbidden_kwargs:
-            raise TypeError(
-                "with_options() got an unexpected keyword argument '%s'" % forbidden_kwargs.pop())
-
-        for key, value in iteritems(kwargs):
+    def with_options(
+            self, codec_options=None, read_preference=None, write_concern=None, read_concern=None):
+        for key, options in iteritems(_WITH_OPTIONS_KWARGS):
+            value = locals()[key]
             if value is None:
                 continue
-            options = _WITH_OPTIONS_KWARGS[key]
             for attr in options.attrs:
                 if not hasattr(value, attr):
                     raise TypeError(
@@ -1606,16 +1642,14 @@ class Collection(object):
                     '%s is a valid parameter for with_options but it is currently not implemented '
                     'in Mongomock' % key)
 
-        write_concern = kwargs.get('write_concern', self.write_concern)
-        read_preference = kwargs.get('read_preference', self.read_preference)
         return Collection(
-            self.database, write_concern=write_concern, read_preference=read_preference,
-            _store=self._store)
+            self.database, self.name, write_concern=write_concern, read_preference=read_preference,
+            _db_store=self._db_store)
 
     def rename(self, new_name, session=None, **kwargs):
         if session:
             raise NotImplementedError('Mongomock does not handle sessions yet')
-        self.database.rename_collection(self.name, new_name, **kwargs)
+        return self.database.rename_collection(self.name, new_name, **kwargs)
 
     def bulk_write(self, requests, ordered=True, bypass_document_validation=False, session=None):
         if bypass_document_validation:
@@ -1694,7 +1728,10 @@ class Cursor(object):
         self._emitted = 0
 
     def sort(self, key_or_list, direction=None):
-        self._sort = helpers.create_index_list(key_or_list, direction)
+        sort = helpers.create_index_list(key_or_list, direction)
+        if not sort:
+            raise ValueError('key_or_list must not be the empty list')
+        self._sort = sort
         self._factory = functools.partial(
             self.collection._get_dataset, self._spec, self._sort, self._projection, dict)
         return self
@@ -1771,12 +1808,11 @@ class Cursor(object):
             self._skip = skip
             self._limit = limit
             return self
-        elif not isinstance(index, int):
+        if not isinstance(index, int):
             raise TypeError("index '%s' cannot be applied to Cursor instances" % index)
-        elif index < 0:
+        if index < 0:
             raise IndexError('Cursor instances do not support negativeindices')
-        else:
-            return self._compute_results(with_limit_and_skip=True)[index]
+        return self._compute_results(with_limit_and_skip=True)[index]
 
     def __enter__(self):
         return self
